part of '../lib.dart';

class Invoke {
  final Units _units;
  const Invoke._(this._units);

  T call<T, K extends Object>(T Function() fn, [K? dep]) {
    dep ??= _constDep;
    return _units.next<T, K>(() => _InvokeUnit(fn), dep as K);
  }

  T deferred<T, K extends Object>(
    T Function() fn, {
    K? dep,
    required Future<void> until,
    required T initialValue,
  }) {
    dep ??= _constDep;

    return _units.next<T, K>(
      () => _InvokeUnit(fn, until, initialValue),
      dep as K,
    );
  }

  Result<T> async<T, K extends Object, R>(
    Future<T> Function() fn, {
    void Function()? abort,
    Future<void>? deferredUntil,
    K? dep,
  }) {
    dep ??= _constDep;
    return _units.next<Result<T>, K>(
      () => _InvokeAsyncUnit(fn, abort, deferredUntil),
      dep as K,
    );
  }

  Result<T> afterLayout<T, K extends Object>(
    T Function() fn, {
    K? dep,
  }) {
    dep ??= _constDep;

    final completer = _units._owner.ref(() => Completer<T>(), dep);

    _units._owner.invoke(() {
      SchedulerBinding.instance.addPostFrameCallback((_) {
        completer.value.complete(fn());
      });
    }, dep);

    return _units.next<Result<T>, K>(
      () => _InvokeAsyncUnit(
        () => completer.value.future,
        null,
        completer.value.future,
      ),
      dep as K,
    );
  }
}

class _InvokeUnit<T, K> extends Unit<T, K> {
  late T value;
  final T Function() _callback;

  Future<void>? _deferredUntil;
  T? _initialValue;

  _InvokeUnit(
    this._callback, [
    this._deferredUntil,
    this._initialValue,
  ]);

  void _extractValue() {
    if (_deferredUntil != null) {
      value = _initialValue as T;
      _deferredUntil!.then((_) {
        value = _callback();
        notify();
      });
    } else {
      value = _callback();
    }
  }

  @override
  void init(CTNode node, K dep) {
    super.init(node, dep);
    _extractValue();
  }

  @override
  Unit<T, K> update(
    CTNode node,
    K oldDep,
    K newDep,
    _InvokeUnit<T, K> newUnit,
  ) {
    super.update(node, oldDep, newDep, newUnit);

    if (oldDep == newDep) {
      return this;
    }

    if (_deferredUntil == newUnit._deferredUntil) {
      value = newUnit._callback();
      return this;
    } else {
      _deferredUntil = newUnit._deferredUntil;
      _initialValue = newUnit._initialValue;
      _extractValue();

      return this;
    }
  }

  @override
  T unwrap() {
    return value;
  }
}

class _InvokeAsyncUnit<T, K> extends Unit<Result<T>, K> {
  final Future<T> Function() _fn;
  final void Function()? _abort;
  final Future<void>? _deferredUntil;

  late Result<T> _value;

  static final Future<void> _defaultDeferred = SynchronousFuture(null);

  _InvokeAsyncUnit(this._fn, [this._abort, this._deferredUntil]);

  @override
  void init(CTNode node, K dep) {
    super.init(node, dep);
    _value = Pending<T>();

    (_deferredUntil ?? _defaultDeferred).whenComplete(_load);
  }

  void _load() {
    _value = Loading<T>();
    notify();

    _fn().then((value) {
      _value = Success(value);
      notify();
    }).catchError((e, s) {
      _value = Failure(e, s);
      notify();
    });
  }

  // TODO: test if this could be dropped
  @override
  Unit<Result<T>, K> update(
    CTNode node,
    K oldDep,
    K newDep,
    _InvokeAsyncUnit<T, K> newUnit,
  ) {
    super.update(node, oldDep, newDep, newUnit);

    if (oldDep == newDep) {
      return this;
    }

    _abort?.call();
    (newUnit._deferredUntil ?? _defaultDeferred).whenComplete(_load);

    return this;
  }

  @override
  void dispose() {
    _abort?.call();
    super.dispose();
  }

  @override
  Result<T> unwrap() {
    return _value;
  }
}
